package starzi.kvf;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.Deque;
import java.util.LinkedList;
import java.util.List;
import java.util.Scanner;

/**
 * This class can be used to parse .kvf files to create KVItem hierarchies.
 * 
 * @author Nic Starzi
 *
 */
public class KVFParser {
	
	/**
	 * Helper method to read a file as a single string.
	 * @param file
	 * @return
	 * @throws FileNotFoundException
	 */
	private static String readFileAsString(File file) throws FileNotFoundException {
		Scanner scanner = new Scanner(file);
		scanner.useDelimiter("\\Z");
		String docAsString = scanner.next();
		scanner.close();
		scanner = null;
		return docAsString;
	}
	
	/** The entire document as String **/
	private final String doc;
	/** Used to parse text **/
	private final StringBuilder sb;
	/** Stack holds items that children will be added to **/
	private final Deque<KVItem> itemStack;
	private int readPos;
	/** The item which is currently being parsed. 
	 * Key-Value-Pairs that are read will be added to this item **/
	private KVItem next;
	
	/**
	 * Creates a new KVFParser for the given document.
	 * @param docAsString
	 */
	public KVFParser(String docAsString) {
		this.doc = docAsString;
		sb = new StringBuilder();
		itemStack = new LinkedList<>();
		readPos = 0;
	}
	
	public KVFParser(File docAsFile) throws FileNotFoundException {
		this(readFileAsString(docAsFile));
	}
	
	/**
	 * Starts the parsing process to create a KVItem hierarchy from the document.
	 * This method will return once the entire document was parsed and is without errors.
	 * 
	 * If the document is not following the .kvf standards then 
	 * this method will throw an appropriate runtime exception.
	 * @return The document item (an empty item) that encapsulates all items within the document.
	 */
	public KVItem parse() {
		itemStack.clear();
		readPos = 0;
		
		KVItem document = new KVItem();
		itemStack.push(document);
		
		while (!isEOF()) {
			readItem();
		}
		
		if (itemStack.peek() != document) {
			throw new IllegalStateException("Illegal number of curly braces.");
		}
		
		return document;
	}
	
	/*
	 * Item creation methods
	 */
	
	/**
	 * Creates a new item instance for 'next'.
	 * Reads an item from the document and populates next with the contents.
	 * Adds next to the current top item on the itemStack and closes items 
	 * from the stack if closing curly brackets are read.
	 */
	private void readItem() {
		// Skip until we find first item
		skipAhead();
		// Create item
		next = new KVItem();
		// Start of Item Declaration
		readStartSymbol();
		// All key-value pairs
		readAllKeyValuePairs();
		// End of Item Declaration
		readEndSymbol();
		
		// This item is a child of the item on top of the stack.
		itemStack.peek().addChild(next);
		
		if (isEOF()) {
			return;
		}
		
		if (currentChar() == '{') {
			/*
			 * Start of a child block. Add to top of stack.
			 */
			readPos++;
			skipAhead();
			
			itemStack.push(next);
		}
		while (!isEOF() && currentChar() == '}') {
			/*
			 * End of a child block. Close the upper most item on the stack.
			 */
			readPos++;
			skipAhead();
			
			/*
			 * Optimization of the children list in an item.
			 * If its an arraylist it is trimmed to size.
			 */
			List<KVItem> itemChildren = itemStack.peek().children;
			if (itemChildren != null && itemChildren instanceof ArrayList) {
				((ArrayList<KVItem>) itemChildren).trimToSize();
			}
			
			itemStack.pop();
		}
	}
	
	/**
	 * Reads Key-Value-Pairs until the '>' character is found.
	 */
	private void readAllKeyValuePairs() {
		// We stop reading kvp's when the item declaration is over.
		while (currentChar() != '>') {
			readKeyValuePair();
		}
	}
	
	/**
	 * Reads one Key-Value-Pair which consists of either a lonely Key, or a Key = "Value" pairing.
	 * The Key-Value-Pair will be added to the 'next' item.
	 */
	private void readKeyValuePair() {
		clearStringBuilder();
		
		// Read key (keys consist only of alphabetical letters)
		while (isAlphabetic()) {
			sb.append(currentChar());
			readPos++;
		}
		
		String key = sb.toString();
		String value = "";
		
		skipAhead();
		// Read equality sign
		if (currentChar() == '=') {
			readPos++;
			skipAhead();
			// Read value
			value = readValue();
		}
		
		if (value.isEmpty()) {
			next.add(key);
		} else {
			next.set(key, value);
		}
	}
	
	/**
	 * Reads a string that is encapsulated within quotation marks.
	 * Advances the 'readPos' behind the hindmost quotation mark.
	 * @return
	 */
	private String readValue() {
		clearStringBuilder();
		// Value is started after quotation mark
		readQuotationMark();
		// Value is terminated by quotation mark
		while (currentChar() != '"') {
			sb.append(currentChar());
			readPos++;
		}
		// to move the readPos past the quotation mark
		readQuotationMark();
		skipAhead();
		
		return sb.toString();
	}
	
	/*
	 * Read Methods (manipulate read pos)
	 */
	
	/**
	 * Attempts to read the "item declaration end symbol" '>'.
	 * If the end symbol can not be read at the current position an exception will be thrown.
	 */
	private void readEndSymbol() {
		if (currentChar() != '>') {
			throwException("Item must end with '>' symbol.");
		}
		readPos++;
		skipAhead();
	}
	
	/**
	 * Attempts to read the "item declaration start symbol" '<'.
	 * If the start symbol can not be read at the current position an exception will be thrown.
	 */
	private void readStartSymbol() {
		if (currentChar() != '<') {
			throwException("Item must start with '<' symbol.");
		}
		readPos++;
		skipAhead();
	}
	
	/**
	 * Attempts to read a quotation mark.
	 * If the quotation mark can not be read at the current position an exception will be thrown.
	 */
	private void readQuotationMark() {
		if (currentChar() != '"') {
			throwException("Value must start with '\"' symbol.");
		}
		readPos++;
	}
	
	/**
	 * Skips all white spaces and comments. 
	 */
	private void skipAhead() {
		skipWhitespaces();
		skipComments();
		skipWhitespaces();
		// TODO: What happens if 2 comments follow each other?
	}
	
	/**
	 * Skips all white spaces.
	 */
	private void skipWhitespaces() {
		if (isEOF()) {
			return;
		}
		while (!isEOF() && Character.isWhitespace(currentChar())) {
			readPos++;
		}
		skipComments();
	}
	
	/**
	 * Skips one entire comment.
	 */
	private void skipComments() {
		if (isEOF()) {
			return;
		}
		if (currentChar() == '/') {
			readPos++;
			while (currentChar() != '/') {
				readPos++;
			}
			readPos++;
		}
	}
	
	/*
	 * Convenience Methods
	 */
	
	/**
	 * Returns the char at the current 'readPos'
	 * If the end of file is reached this method will throw an exception.
	 * @return
	 */
	private char currentChar() {
		if (isEOF()) {
			throwException("End of File reached.");
		}
		return doc.charAt(readPos);
	}
	
	/**
	 * Returns true if the char at the current 'readPos' is alphabetic.
	 * @return
	 */
	private boolean isAlphabetic() {
		char c = currentChar();
		return Character.isAlphabetic(c);
	}
	
	/**
	 * Returns true if the end of file is reached.
	 * @return
	 */
	private boolean isEOF() {
		return readPos >= doc.length();
	}
	
	/**
	 * Deletes the contents of the stringbuilder.
	 */
	private void clearStringBuilder() {
		sb.delete(0, sb.length());
	}
	
	/**
	 * Throws a runtime exception.
	 * A needless convenience method.
	 * @param msg
	 */
	private void throwException(String msg) {
		throw new RuntimeException(msg);
	}
	
}