package starzi.kvf;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

/**
 * An item within a .kvf hierarchy.
 * Items can have any number of other items as children. 
 * If an item is the child of an other item then the item has a parent.
 * Items can have any number of Key-Value-Pairs as values.
 * 
 * There is a "missing item" instance accessible via a public static variable called MISSING_ITEM.
 * This instance is returned if the user attempts to access a KVItem that does not exist.
 * Checking whether an item is the MISSING_ITEM can either be done by comparing 
 * with the public static variable, or by calling the method exists() on the instance.
 * 
 * @author Nic Starzi
 *
 */
public class KVItem implements Cloneable {
	
	/** This special instance of KVItem acts as a placeholder for all non-existent items in a .kvf hierarchy.
	 * You can compare an item with the MISSING_ITEM to check whether that item exists or not. */
	public static final KVItem MISSING_ITEM = MissingItem.INSTANCE;
	
	/** The list of children is lazily initialized only if children are added. */
	protected List<KVItem>	children;
	/** The list of Key-Value-Pairs is lazily initialized only if KVPairs are added. */
	protected List<KVPair>	values;
	/** This is null if this item has no parent. 
	 * If this item is the child of an other item then this variable should never be null. */
	protected KVItem		parent;
	
	/**
	 * Constructs a new KVItem without children, values or a parent.
	 */
	public KVItem() {
		children = null;
		values = null;
		parent = null;
	}
	
	/**
	 * Returns true if this item exists and is not a "missing item".
	 * When this method returns false then the user attempted to traverse the .kvf hierarchy on a non-existent path.
	 * @return
	 */
	public boolean exists() {
		return true;
	}
	
	/*
	 * Tree Structure Code
	 */
	
	/**
	 * The given argument will become the parent of this item, likewise this item will become a child of the given argument.
	 * If this item already has a parent then it will first be removed from its parent and no longer be a child of it.
	 * If the argument is the "missing item" then this method will throw an IllegalArgumentException.
	 * If the argument is null then this item will no longer have a parent.
	 * @param item
	 * @throws IllegalArgumentException if the "missing item" is given as argument.
	 * @throws UnsupportedOperationException if this is the "missing item".
	 */
	public void setParent(KVItem item) {
		if (item == MISSING_ITEM) {
			throw new IllegalArgumentException("Parent can not be the missing item.");
		}
		if (hasParent()) {
			getParent().removeChild(this);
		}
		if (item != null) {
			item.addChild(this);
		}
	}
	
	/**
	 * Returns either the parent of this item, if it has a parent, or the "missing item" if it does not have a parent.
	 * If this item does not exist, that is, if this item is the "missing item", then the "missing item" is returned.
	 * @return
	 */
	public KVItem getParent() {
		if (parent == null) {
			return MISSING_ITEM;
		}
		return parent;
	}
	
	/**
	 * Returns true if this item has a parent that exists, that is, the parent is not the "missing item".
	 * @return
	 */
	public boolean hasParent() {
		return getParent() != MISSING_ITEM;
	}
	
	/**
	 * Adds the argument as a child to this item.
	 * If the argument is null then a NullPointerException will be thrown since items may never be null.
	 * If the argument is the "missing item" then an IllegalArgumentException will be thrown.
	 * If the argument has a parent then an IllegalStateException will be thrown because an item can not have 2 parents.
	 * 
	 * If none of the above situations take place then this item will become the parent of the argument and the argument will become a child of this item.
	 * @param item
	 * @throws NullPointerException if argument is null
	 * @throws IllegalArgumentException for the "missing item"
	 * @throws IllegalStateException if argument already has a parent
	 * @throws UnsupportedOperationException if this is the "missing item".
	 */
	public void addChild(KVItem item) {
		if (item == null) {
			throw new NullPointerException();
		}
		if (item == MISSING_ITEM) {
			throw new IllegalArgumentException("Missing Item");
		}
		if (item.hasParent()) {
			throw new IllegalStateException();
		}
		item.parent = this;
		lazyInitChildren();
		children.add(item);
	}
	
	/**
	 * Removes the argument from the children of this item.
	 * If this item is not the parent of the argument then an IllegalStateException will be thrown.
	 * If the argument is the "missing item" then an IllegalArgumentException will be thrown.
	 * If the argument is null a NullPointerException will be thrown.
	 * 
	 * If none of the above situations take place then the argument will not have a parent and not be one of the children of this item anymore.
	 * @param item
	 * @throws NullPointerException if argument is null
	 * @throws IllegalArgumentException for the "missing item"
	 * @throws IllegalStateException if this is not the parent of the argument
	 * @throws UnsupportedOperationException if this is the "missing item".
	 */
	public void removeChild(KVItem item) {
		if (!this.equals(item.getParent())) {
			throw new IllegalStateException();
		}
		if (item == MISSING_ITEM) {
			throw new IllegalArgumentException("Missing Item");
		}
		item.parent = null;
		children.remove(item);
	}
	
	/**
	 * Returns the first child which has the given key as a key to one of its values.
	 * Please note that the key is not case-sensitive.
	 * If there is no such child then the "missing item" will be returned.
	 * 
	 * This method should never return null and never throw an exception.
	 * @param key
	 * @return
	 */
	public KVItem getChildWithKey(String key) {
		if (children == null) {
			return MISSING_ITEM;
		}
		for (KVItem child : children) {
			if (child.has(key)) {
				return child;
			}
		}
		return MISSING_ITEM;
	}
	
	/**
	 * Returns the first child which has the given key mapped to the given value.
	 * Please note that neither key nor value are case-sensitive.
	 * If there is no such child then the "missing item" will be returned.
	 * 
	 * This method should never return null and never throw an exception.
	 * @param key
	 * @return
	 */
	public KVItem getChildWithValue(String key, String value) {
		if (children == null) {
			return MISSING_ITEM;
		}
		for (KVItem child : getChildren()) {
			if (child.has(key, value)) {
				return child;
			}
		}
		return MISSING_ITEM;
	}
	
	/**
	 * Returns a list of KVItem instances where each element in the list has the given key as a key to one of its values.
	 * Please note that the key is not case-sensitive.
	 * If there are no such items then an empty list will be returned.
	 * 
	 * No assumptions should be made whether the returned list is modifiable or not.
	 * 
	 * This method should never return null and never throw an exception.
	 * @param key
	 * @return
	 */
	public List<KVItem> getChildrenWithKey(String key) {
		List<KVItem> list = null;
		for (KVItem child : getChildren()) {
			if (child.has(key)) {
				// Lazy initialization of the list.
				if (list == null) {
					list = new ArrayList<>();
				}
				list.add(child);
			}
		}
		if (list == null) {
			list = Collections.emptyList();
		}
		return list;
	}
	
	/**
	 * Returns a list of KVItem instances where each element in the list has the given key mapped to the given value.
	 * Please note that neither key nor value are case-sensitive.
	 * If there are no such items then an empty list will be returned.
	 * 
	 * No assumptions should be made whether the returned list is modifiable or not.
	 * 
	 * This method should never return null and never throw an exception.
	 * @param key
	 * @return
	 */
	public List<KVItem> getChildrenWithValue(String key, String value) {
		List<KVItem> list = null;
		for (KVItem child : getChildren()) {
			if (child.has(key, value)) {
				// Lazy initialization of the list.
				if (list == null) {
					list = new ArrayList<>();
				}
				list.add(child);
			}
		}
		if (list == null) {
			list = Collections.emptyList();
		}
		return list;
	}
	
	/**
	 * If the index is out of bounds then an IndexOutOfBoundsException will be thrown.
	 * This method never returns null nor the "missing item". All items within the range of valid indices are actual items that exist.
	 * 
	 * The lower bound for indices is 0, the upper bound can be retrieved with the method getChildCount().
	 * 
	 * The indices for the children are in the order the children were added to this item.
	 * @param index
	 * @return
	 * @throws IndexOutOfBoundsException
	 * @throws UnsupportedOperationException if this is the "missing item".
	 */
	public KVItem getChild(int index) {
		if (index < 0 || index >= getChildCount()) {
			throw new IndexOutOfBoundsException(index+" must be between 0 and "+getChildCount());
		}
		return children.get(index);
	}
	
	/**
	 * Returns true if this item has at least 1 child that is not the "missing item".
	 * @return
	 */
	public boolean hasChildren() {
		return getChildCount() > 0;
	}
	
	/**
	 * Returns the number of children that this item has.
	 * @return
	 */
	public int getChildCount() {
		if (children == null) {
			return 0;
		}
		return children.size();
	}
	
	/**
	 * Returns an unmodifiable collection of all the children of this item.
	 * The collection will never contain any null elements or non-existent items.
	 * 
	 * This method will never return null and never throw an exception.
	 * @return
	 */
	public Collection<KVItem> getChildren() {
		if (children == null) {
			return Collections.emptyList();
		}
		return Collections.unmodifiableCollection(children);
	}
	
	/*
	 * Key-Value Code
	 */
	
	/**
	 * Constructs a new Key-Value-Pair from the given key.
	 * The value for the key will be the empty string.
	 * If there is already a Key-Value-Pair with the same key then its value will be overwritten with the empty string.
	 * 
	 * This is a convenience method and equal to:
	 * set(key, "")
	 * @param key
	 * @throws UnsupportedOperationException if this is the "missing item".
	 */
	public void add(String key) {
		set(key, "");
	}
	
	/**
	 * Constructs a new Key-Value-Pair from the given arguments.
	 * If there is already a Key-Value-Pair with the same key then its value will be overwritten.
	 * @param key
	 * @throws UnsupportedOperationException if this is the "missing item".
	 */
	public void set(String key, String value) {
		lazyInitValues();
		int index = getIndex(key);
		if (index != -1) {
			values.remove(index);
		}
		values.add(new KVPair(key, value));
	}
	
	/**
	 * Returns true if this item has a Key-Value-Pair with the given key.
	 * If this item is the "missing item" then this method will always return false.
	 * @param key
	 * @return
	 */
	public boolean has(String key) {
		return get(key) != KVPair.MISSING_PAIR;
	}
	
	/**
	 * Returns true if this item has a Key-Value-Pair with the given key and value.
	 * If this item is the "missing item" then this method will always return false.
	 * @param key
	 * @return
	 */
	public boolean has(String key, String value) {
		KVPair pair = get(key);
		if (pair != KVPair.MISSING_PAIR) {
			return pair.asString().equalsIgnoreCase(value);
		}
		return false;
	}
	
	/**
	 * Returns the first Key-Value-Pair instance that has the given key.
	 * If there is no such pair then the "missing pair" is returned.
	 * If this item is the "missing item" then the "missing pair" will always be returned.
	 * 
	 * This method will never return null or throw an exception.
	 * @param key
	 * @return
	 */
	public KVPair get(String key) {
		if (values == null) {
			return KVPair.MISSING_PAIR;
		}
		int index = getIndex(key);
		if (index == -1) {
			return KVPair.MISSING_PAIR;
		}
		return values.get(index);
	}
	
	/**
	 * Removes the Key-Value-Pair that is mapped to the given key.
	 * If there is no such pair then nothing will happen.
	 * @param key
	 * @throws UnsupportedOperationException if this is the "missing item".
	 */
	public void remove(String key) {
		if (values == null) {
			return;
		}
		int index = getIndex(key);
		if (index != -1) {
			values.remove(index);
		}
	}
	
	/**
	 * Returns an unmodifiable collection of all the Key-Value-Pairs within this item.
	 * The collection will never contain any null elements or non-existent pairs.
	 * 
	 * This method will never return null and never throw an exception.
	 * @return
	 */
	public Collection<KVPair> getValues() {
		if (values == null) {
			return Collections.emptyList();
		}
		return Collections.unmodifiableCollection(values);
	}
	
	/**
	 * Convenience helper method to get the index of a pair with the given key.
	 * Returns -1 if the pair is not contained.
	 * @param key
	 * @return
	 */
	private int getIndex(String key) {
		for (int i = 0; i < values.size(); i++) {
			KVPair pair = values.get(i);
			if (pair.getKey().equalsIgnoreCase(key)) {
				return i;
			}
		}
		return -1;
	}
	
	/*
	 * Maintenance Code
	 */
	
	/**
	 * Returns only this item with all of its Key-Value-Pairs in a human readable form.
	 * If you want to have the entire sub-hierarchy use the deepToString() method instead.
	 * @return
	 */
	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append('<');
		for (KVPair pair : getValues()) {
			sb.append(' ');
			sb.append(pair.getKey());
			
			if (!pair.asString().isEmpty()) {
				sb.append('=');
				sb.append('"');
				sb.append(pair.asString());
				sb.append('"');
			}
		}
		sb.append(' ');
		sb.append('>');
		if (hasChildren()) {
			sb.append(" { ... }");
		}
		return sb.toString();
	}
	
	/**
	 * Returns the entire sub-hierarchy defined by this item in a human readable form.
	 * This method will return a string that represents this hierarchy in clear text format.
	 * This string can be written to a file to create a valid .kvf file.
	 * @return
	 */
	public String deepToString() {
		StringBuilder sb = new StringBuilder();
		writeToStringBuilder(sb, 0);
		return sb.toString();
	}
	
	/**
	 * Convenience helper method used by the deepToString() method.
	 * @param sb
	 * @param tabCount
	 */
	private void writeToStringBuilder(StringBuilder sb, int tabCount) {
		for (int i = 0; i < tabCount; i++) {
			sb.append('\t');
		}
		
		sb.append('<');
		for (KVPair pair : getValues()) {
			sb.append(' ');
			sb.append(pair.getKey());
			
			if (!pair.asString().isEmpty()) {
				sb.append('=');
				sb.append('"');
				sb.append(pair.asString());
				sb.append('"');
			}
		}
		sb.append(' ');
		sb.append('>');
		if (hasChildren()) {
			sb.append(' ');
			sb.append('{');
		}
		
		if (hasChildren()) {
			for (KVItem child : getChildren()) {
				sb.append('\r');
				sb.append('\n');
				child.writeToStringBuilder(sb, tabCount + 1);
			}
			sb.append('\r');
			sb.append('\n');
			for (int i = 0; i < tabCount; i++) {
				sb.append('\t');
			}
			sb.append('}');
		}
	}
	
	/**
	 * Opens the given file and attempts to write the sub-hierarchy defined by this item to the file.
	 * If the file already exists it will be overwritten by this method.
	 * If the file does not exist it will be created.
	 * 
	 * The written file will be a valid .kvf file which can be parsed to recreate this sub-hierarchy.
	 * @param file
	 * @throws FileNotFoundException
	 */
	public void writeToFile(File file) throws FileNotFoundException {
		try (PrintWriter writer = new PrintWriter(file)) {
			writer.println(deepToString());
		}
	}
	
	/**
	 * Makes a deep copy of this item, all of its children and all of its Key-Value-Pairs.
	 */
	public KVItem clone() {
		KVItem clone = new KVItem();
		
		for (KVItem child : getChildren()) {
			KVItem clonedChild = child.clone();
			clonedChild.parent = null;
			clone.addChild(clonedChild);
		}
		
		for (KVPair pair : getValues()) {
			clone.set(pair.getKey(), pair.asString());
		}
		
		return clone;
	}
	
	/*
	 * Lazy Initialization Methods
	 */
	
	/** What it says on the tin */
	private void lazyInitChildren() {
		if (children == null) {
			children = new ArrayList<>();
		}
	}
	
	/** What it says on the tin */
	private void lazyInitValues() {
		if (values == null) {
			values = new ArrayList<>();
		}
	}
	
}