package starzi.kvf;

/**
 * A pair of 2 Strings, one of which is a key to the other value.
 * This class provides methods for convenient value transformation for various often used primitive data types.
 * 
 * KVPairs are used within KVItems in a .kvf hierarchy.
 * 
 * @author Nic Starzi
 *
 */
public class KVPair implements Cloneable {
	
	/** This special instance of KVPair acts as a placeholder for all non-existent pairs in an item.
	 * You can compare a pair with the MISSING_PAIR to check whether that item exists or not. */
	public static final KVPair MISSING_PAIR = MissingPair.INSTANCE;
	
	private final String key;
	private final String value;
	
	/**
	 * Constructs a new Key-Value-Pair from the given arguments.
	 * Neither key nor value can be null, otherwise an exception will be thrown.
	 * @param key
	 * @param value
	 * @throws NullPointerException
	 */
	protected KVPair(String key, String value) {
		if (key == null || value == null) {
			throw new NullPointerException();
		}
		this.key = key;
		this.value = value;
	}
	
	/**
	 * This method returns true if this pair actually exists within the .kvf hierarchy.
	 * If this method returns false then this pair is the "missing pair".
	 * @return
	 */
	public boolean exists() {
		return true;
	}
	
	/**
	 * Returns the key of this KVPair as String.
	 * 
	 * This should never be null nor throw any exceptions.
	 * @return
	 */
	public String getKey() {
		return key;
	}
	
	/**
	 * Returns the value of this pair as String.
	 * If the key had no value associated within the document then this method will return an empty String.
	 * 
	 * This method will never return null nor throw any exceptions.
	 * @return
	 */
	public String asString() {
		return value;
	}
	
	/**
	 * Returns the value of this pair as String if the value is not the empty string.
	 * Otherwise the defaultValue will be returned.
	 * 
	 * This method will never return null nor throw any exceptions.
	 * @param defaultValue
	 * @return
	 */
	public String asString(String defaultValue) {
		if (asString().isEmpty()) {
			return defaultValue;
		}
		return asString();
	}
	
	/**
	 * Returns the value of this pair as a boolean.
	 * More specifically, if the value is equal to "true", ignoring capitalization, then this method will return true.
	 * If the value is equal to "false", ignoring capitalization, then this method will return false.
	 * 
	 * In any other situation this method will throw a FormatException.
	 * @return
	 * @throws FormatException if value as string is neither "true" nor "false"
	 */
	public boolean asBoolean() throws FormatException {
		if (asString().equalsIgnoreCase("true")) {
			return true;
		} else if (asString().equalsIgnoreCase("false")) {
			return false;
		}
		throw new FormatException(asString()+" can not be parsed to boolean.");
	}
	
	/**
	 * If the value of this pair can be returned as a boolean then that will be returned.
	 * Otherwise the defaultValue will be returned.
	 * 
	 * This method will never throw any exceptions.
	 * @param defaultValue
	 * @return
	 */
	public boolean asBoolean(boolean defaultValue) {
		try {
			return asBoolean();
		} catch (FormatException e) {}
		return defaultValue;
	}
	
	/**
	 * Returns the value of this pair as a byte.
	 * If this value can not be parsed into a byte a FormatException will be thrown.
	 * @return
	 * @throws FormatException if the value can not be parsed
	 */
	public byte asByte() {
		try {
			return Byte.parseByte(asString());
		} catch (NumberFormatException e) {}
		throw new FormatException(asString()+" can not be parsed to byte.");
	}
	
	/**
	 * If the value of this pair can be returned as a byte then that will be returned.
	 * Otherwise the defaultValue will be returned.
	 * 
	 * This method will never throw any exceptions.
	 * @param defaultValue
	 * @return
	 */
	public byte asByte(byte defaultValue) {
		try {
			return asByte();
		} catch (FormatException e) {}
		return defaultValue;
	}
	
	/**
	 * Returns the value of this pair as a short.
	 * If this value can not be parsed into a short a FormatException will be thrown.
	 * @return
	 * @throws FormatException if the value can not be parsed
	 */
	public short asShort() {
		try {
			return Short.parseShort(asString());
		} catch (NumberFormatException e) {}
		throw new FormatException(asString()+" can not be parsed to short.");
	}
	
	/**
	 * If the value of this pair can be returned as a short then that will be returned.
	 * Otherwise the defaultValue will be returned.
	 * 
	 * This method will never throw any exceptions.
	 * @param defaultValue
	 * @return
	 */
	public short asShort(short defaultValue) {
		try {
			return asShort();
		} catch (FormatException e) {}
		return defaultValue;
	}
	
	/**
	 * Returns the value of this pair as an int.
	 * If this value can not be parsed into an int a FormatException will be thrown.
	 * @return
	 * @throws FormatException if the value can not be parsed
	 */
	public int asInt() {
		try {
			return Integer.parseInt(asString());
		} catch (NumberFormatException e) {}
		throw new FormatException(asString()+" can not be parsed to int.");
	}
	
	/**
	 * If the value of this pair can be returned as an int then that will be returned.
	 * Otherwise the defaultValue will be returned.
	 * 
	 * This method will never throw any exceptions.
	 * @param defaultValue
	 * @return
	 */
	public int asInt(int defaultValue) {
		try {
			return asInt();
		} catch (FormatException e) {}
		return defaultValue;
	}
	
	/**
	 * Returns the value of this pair as a long.
	 * If this value can not be parsed into a long a FormatException will be thrown.
	 * @return
	 * @throws FormatException if the value can not be parsed
	 */
	public long asLong() {
		try {
			return Long.parseLong(asString());
		} catch (NumberFormatException e) {}
		throw new FormatException(asString()+" can not be parsed to long.");
	}
	
	/**
	 * If the value of this pair can be returned as a long then that will be returned.
	 * Otherwise the defaultValue will be returned.
	 * 
	 * This method will never throw any exceptions.
	 * @param defaultValue
	 * @return
	 */
	public long asLong(long defaultValue) {
		try {
			return asLong();
		} catch (FormatException e) {}
		return defaultValue;
	}
	
	/**
	 * Returns the value of this pair as a float.
	 * If this value can not be parsed into a float a FormatException will be thrown.
	 * @return
	 * @throws FormatException if the value can not be parsed
	 */
	public float asFloat() {
		try {
			return Float.parseFloat(asString());
		} catch (NumberFormatException e) {}
		throw new FormatException(asString()+" can not be parsed to float.");
	}
	
	/**
	 * If the value of this pair can be returned as a float then that will be returned.
	 * Otherwise the defaultValue will be returned.
	 * 
	 * This method will never throw any exceptions.
	 * @param defaultValue
	 * @return
	 */
	public float asFloat(float defaultValue) {
		try {
			return asFloat();
		} catch (FormatException e) {}
		return defaultValue;
	}
	
	/**
	 * Returns the value of this pair as a double.
	 * If this value can not be parsed into a double a FormatException will be thrown.
	 * @return
	 * @throws FormatException if the value can not be parsed
	 */
	public double asDouble() {
		try {
			return Double.parseDouble(asString());
		} catch (NumberFormatException e) {}
		throw new FormatException(asString()+" can not be parsed to double.");
	}
	
	/**
	 * If the value of this pair can be returned as a double then that will be returned.
	 * Otherwise the defaultValue will be returned.
	 * 
	 * This method will never throw any exceptions.
	 * @param defaultValue
	 * @return
	 */
	public double asDouble(double defaultValue) {
		try {
			return asDouble();
		} catch (FormatException e) {}
		return defaultValue;
	}
	
	/**
	 * This method will try to parse the value of this pair into an enum of the same type as the given defaultValue.
	 * If the parsing fails because the value does not represent any enum of that type then the defaultValue will be returned.
	 * @param defaultValue
	 * @throws NullPointerException when defaultValue is null.
	 * @return
	 */
	public Enum<?> asEnum(Enum<?> defaultValue) {
		for (Enum<?> e : defaultValue.getDeclaringClass().getEnumConstants()) {
			if (e.name().equalsIgnoreCase(asString())) {
				return e;
			}
		}
		return defaultValue;
	}
	
	public Enum<?> asEnum(Class<Enum<?>> enumClass) {
		for (Enum<?> e : enumClass.getEnumConstants()) {
			if (e.name().equalsIgnoreCase(asString())) {
				return e;
			}
		}
		throw new FormatException(asString()+" can not be parsed into "+enumClass.getSimpleName());
	}
	
	/**
	 * Returns this pair in the following format:
	 * < getKey() := getValue() >
	 */
	public String toString() {
		return "< "+getKey()+" := "+asString()+" >";
	}
	
	/**
	 * Returns an exact clone of this pair.
	 */
	public KVPair clone() {
		return new KVPair(getKey(), asString());
	}
	
	/**
	 * Returns true if the value of this KVPair is equal to the given object.
	 * This is done for convenience.
	 */
	public boolean equals(Object obj) {
		return asString().equals(obj);
	}
	
}