package starzi.kvf.tests;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.List;

import starzi.kvf.KVFParser;
import starzi.kvf.KVItem;
import starzi.kvf.KVPair;

public class TestHeroes {
	
	public static void main(String[] args) throws FileNotFoundException {
		KVFParser parser = new KVFParser(new File("Documents/Heroes.kvf"));
		KVItem party = parser.parse();
		
		List<Hero> heroes = new ArrayList<>();
		for (KVItem heroAsItem : party.getChildrenWithKey("hero")) {
			heroes.add(heroFromItem(heroAsItem));
		}
		
		Hero missingHero = heroFromItem(party.getChildWithValue("name", "Gandalf"));
		heroes.add(missingHero);
		
		for (Hero hero : heroes) {
			System.out.println(hero);
			System.out.println("--------------------");
		}
		
		KVItem rogar = party.getChildWithValue("name","Rogar");
		KVItem clonedRogar = rogar.clone();
		System.out.println("Cloned Rogar (before change)");
		System.out.println(clonedRogar.deepToString());
		
		rogar.remove("title");
		rogar.removeChild(rogar.getChildWithKey("attributes"));
		System.out.println("Rogar (after change)");
		System.out.println(rogar.deepToString());
		System.out.println("Cloned Rogar (after change)");
		System.out.println(clonedRogar.deepToString());
	}
	
	public static Hero heroFromItem(KVItem heroAsItem) {
		Hero heroAsHero = new Hero(heroAsItem.get("Name").asString("NAME"), heroAsItem.get("Title").asString("TITLE"));
		
		AliveState heroState = AliveState.UNKNOWN;
		if (heroAsItem.exists()) {
			if (heroAsItem.getChildWithKey("dead").exists()) {
				heroState = AliveState.DEAD;
			} else {
				heroState = AliveState.ALIVE;
			}
		}
		heroAsHero.setState(heroState);
		
		int DEFAULT_ATTRIBUTE_VALUE = 0;
		KVItem attributes = heroAsItem.getChildWithKey("attributes");
		for (Attribute attr : Attribute.values()) {
			KVPair asPair = attributes.get(attr.name());
			int value = asPair.asInt(DEFAULT_ATTRIBUTE_VALUE);
			
			heroAsHero.setAttribute(attr, value);
		}
		
		KVItem inventory = heroAsItem.getChildWithKey("inventory");
		for (KVItem itemAsKVItem : inventory.getChildrenWithKey("item")) {
			String type = itemAsKVItem.get("type").asString();
			String id = itemAsKVItem.get("id").asString();
			int amount = itemAsKVItem.get("amount").asInt(Item.DEFAULT_AMOUNT);
			
			Item item = new Item(id, type);
			item.setAmount(amount);
			
			heroAsHero.addItem(item);
		}
		
		return heroAsHero;
	}
	
	public static class Hero {
		
		private final String name;
		private final String title;
		private final int[] attributes;
		private final List<Item> items;
		private AliveState state;
		
		public Hero(String name, String title) {
			this.name = name;
			this.title = title;
			this.attributes = new int[Attribute.values().length];
			this.items = new ArrayList<>();
		}
		
		public void setState(AliveState state) {
			this.state = state;
		}
		
		public AliveState getState() {
			return state;
		}
		
		public void setAttribute(Attribute attr, int value) {
			attributes[attr.ordinal()] = value;
		}
		
		public int getAttribute(Attribute attr) {
			return attributes[attr.ordinal()];
		}
		
		public void addItem(Item item) {
			if (item == null) {
				throw new NullPointerException();
			}
			items.add(item);
		}
		
		public List<Item> getItems() {
			return items;
		}
		
		public String toString() {
			return name+" "+title+"\n"
				+ "State: "+getState().toString()+"\n"
				+ "Str: "+getAttribute(Attribute.STR)+"\n"
				+ "Dex: "+getAttribute(Attribute.DEX)+"\n"
				+ "Gle: "+getAttribute(Attribute.GLE)+"\n"
				+ "Luc: "+getAttribute(Attribute.LUC)+"\n"
				+ "Sta: "+getAttribute(Attribute.STA)+"\n"
				+ "Int: "+getAttribute(Attribute.INT)+"\n"
				+ "Chr: "+getAttribute(Attribute.CHR)+"\n"
				+ "Items: "+getItems();
		}
		
	}
	
	public static class Item {
		
		public static final int DEFAULT_AMOUNT = 1;
		
		private String id;
		private String type;
		private int amount;
		
		public Item(String id, String type) {
			this.id = id;
			this.type = type;
			amount = DEFAULT_AMOUNT;
		}
		
		public void setAmount(int amount) {
			if (amount <= 0) {
				throw new IllegalArgumentException();
			}
			this.amount = amount;
		}
		
		public String toString() {
			if (amount == DEFAULT_AMOUNT) {
				return " ("+type+") " + id;
			}
			return " ("+type+") " + id + " ["+amount+"]";
		}
	}
	
	public static enum AliveState {
		ALIVE,
		DEAD,
		UNDEAD,
		UNKNOWN;
	}
	
	public static enum Attribute {
		STR,
		DEX,
		GLE,
		LUC,
		STA,
		INT,
		CHR;
	}
	
}